package org.nrg.xnat.gateway;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.text.DateFormat;
import java.util.Calendar;
import java.util.Date;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Properties;
import java.util.TreeMap;
import java.util.zip.ZipEntry;
import java.util.zip.ZipInputStream;

import org.apache.commons.httpclient.HttpMethodBase;
import org.apache.log4j.BasicConfigurator;
import org.apache.log4j.FileAppender;
import org.apache.log4j.Level;
import org.apache.log4j.Logger;
import org.apache.log4j.Priority;
import org.apache.log4j.SimpleLayout;
import org.apache.log4j.varia.NullAppender;
import org.dom4j.io.SAXReader;
import org.nrg.xnat.desktop.tools.XNATTableParser;
import org.nrg.xnat.desktop.utils.Utils;
import org.nrg.xnat.repository.XNATRestAdapter;

import com.pixelmed.dicom.Attribute;
import com.pixelmed.dicom.AttributeList;
import com.pixelmed.dicom.AttributeTag;
import com.pixelmed.dicom.AttributeTagAttribute;
import com.pixelmed.dicom.DicomException;
import com.pixelmed.dicom.DicomInputStream;
import com.pixelmed.dicom.FileMetaInformation;
import com.pixelmed.dicom.InformationEntity;
import com.pixelmed.dicom.SetOfDicomFiles;
import com.pixelmed.dicom.StoredFilePathStrategy;
import com.pixelmed.dicom.TagFromName;
import com.pixelmed.dicom.TransferSyntax;
import com.pixelmed.network.DicomNetworkException;
import com.pixelmed.network.NetworkApplicationInformationFederated;
import com.pixelmed.network.NetworkApplicationProperties;
import com.pixelmed.network.ResponseStatus;
import com.pixelmed.network.StorageSOPClassSCPDispatcher;
import com.pixelmed.query.QueryResponseGenerator;
import com.pixelmed.query.QueryResponseGeneratorFactory;
import com.pixelmed.query.RetrieveResponseGenerator;
import com.pixelmed.query.RetrieveResponseGeneratorFactory;

public class XNATGatewayServer
{
	private QueryResponseGeneratorFactory m_qrgf=null;
	private RetrieveResponseGeneratorFactory m_rrgf=null;
	protected long m_RandSeed=0;
//	private Properties m_properties;
	protected String m_StoreFolder, m_XNATServer, m_XNATUser, m_XNATPass;
	protected int m_maxCacheFiles=50000;
	public static final String m_ver="Mar 30 2010";
	
	public XNATGatewayServer (Properties p)
	throws DicomNetworkException, IOException
	{
		m_qrgf=new XNATQueryResponseGeneratorFactory();
		m_rrgf=new XNATRetrieveResponseGeneratorFactory();
	
		//init logger
		FileAppender appender=null;
		SimpleLayout layout=new SimpleLayout();
		try
		{
			appender=new FileAppender(layout,"./gateway.log",false);
			m_maxCacheFiles=Integer.parseInt(p.getProperty("Application.FilesInCache"));
		}catch(Exception e){}
		Logger l=Logger.getRootLogger();
		BasicConfigurator.configure(new NullAppender());
		l.addAppender(appender);
		l.setLevel(Level.INFO);
		l.info(DateFormat.getDateTimeInstance().format(new Date())+" Server started");
				
		NetworkApplicationProperties netwAppProps = new NetworkApplicationProperties(p);
		NetworkApplicationInformationFederated fedNetwAppInfo = new NetworkApplicationInformationFederated();
		fedNetwAppInfo.startupAllKnownSourcesAndRegister(netwAppProps);
	
		
		m_StoreFolder=p.getProperty("Application.SavedImagesFolderName");
		File sf=new File(m_StoreFolder);
		if(!sf.exists()) sf.mkdir();
		
		m_XNATServer=p.getProperty(XnatServerProperties.XNATServerURL);
		m_XNATUser=p.getProperty(XnatServerProperties.XNATUser);
		m_XNATPass=p.getProperty(XnatServerProperties.XNATPass);
		
		//clean up from the previous run
		try
		{
			for(File f:new File(m_StoreFolder).listFiles()) f.delete();
		}
		catch(Exception e){}
		{
			int port = netwAppProps.getListeningPort();
			String calledAETitle = netwAppProps.getCalledAETitle();
			int storageSCPDebugLevel = netwAppProps.getStorageSCPDebugLevel();
			int queryDebugLevel = netwAppProps.getQueryDebugLevel();
			File f=new File(m_StoreFolder);
			
			new Thread(new StorageSOPClassSCPDispatcher(port,calledAETitle,f,
					StoredFilePathStrategy.BYSOPINSTANCEUIDHASHSUBFOLDERS,null /*new OurReceivedObjectHandler()*/,
					m_qrgf/*this.databaseInformationModel.getQueryResponseGeneratorFactory(queryDebugLevel)*/,
					m_rrgf/*this.databaseInformationModel.getRetrieveResponseGeneratorFactory(queryDebugLevel)*/,
					fedNetwAppInfo/*federatedNetworkApplicationInformation*/,
					false/*secureTransport*/,
					storageSCPDebugLevel)).start();
		}
	}

	public static void main(String arg[]) 
	{
		String propertiesFileName = arg.length > 0 ? arg[0] : "./config/gateway.properties";
		
		try
		{
			Properties props=new Properties();
			try 
			{
				FileInputStream in = new FileInputStream(propertiesFileName);
				props.load(in);
				in.close();
			}
			catch (IOException e)
			{
				Tools.LogException(Priority.ERROR, "Unable to read properties file", e);
			}
			XNATQueryGenerator.LoadVocabulary("./config/vocabulary.xml");
			System.err.println("XNAT/DICOM gateway server, ver. "+m_ver);
			System.err.println("properties="+props);			
			new XNATGatewayServer(props);
		}
		catch (Exception e) 
		{
			Tools.LogException(Priority.ERROR, "Initialization exception", e);
		}
	}
	
	protected InformationEntity getInformationEntityForQueryRetieveLevel(String queryRetrieveLevel) 
	{
		if (queryRetrieveLevel == null)					return null;
		else if (queryRetrieveLevel.equals("PATIENT"))	return InformationEntity.PATIENT;
		else if (queryRetrieveLevel.equals("STUDY"))	return InformationEntity.STUDY;
		else if (queryRetrieveLevel.equals("SERIES"))	return InformationEntity.SERIES;
		else if (queryRetrieveLevel.equals("IMAGE"))	return InformationEntity.INSTANCE;
		else											return null;
	}

	private class XNATRetrieveResponseGeneratorFactory implements RetrieveResponseGeneratorFactory
	{
		private LinkedList<String> m_fileCache=new LinkedList<String>();
		public void UpdateCache(SetOfDicomFiles sdf)
		{
//			final int nMaxFiles=50000;
			int add_sz=sdf.size(),
			    prev_sz=m_fileCache.size();
			int new_sz=m_fileCache.size();
	
			SetOfDicomFiles.DicomFile next;
			for(Iterator it=sdf.iterator(); it.hasNext();)
			{
				next=(SetOfDicomFiles.DicomFile)it.next();
				m_fileCache.add(next.getFileName());
			}
			//in case when there was added more than there was in cache already, return
			if(add_sz>prev_sz) return;
			int to_delete=new_sz-m_maxCacheFiles;
			if(add_sz>m_maxCacheFiles) to_delete=new_sz-add_sz;
			//in case when cache is not completely full, return			
			if(to_delete<=0) return;
			String fname;
			for(int i=0; i<to_delete; i++)
			{
				fname=m_fileCache.getFirst();
				new File(fname).delete();
				m_fileCache.removeFirst();
			}			
		}
		class TestRetrieveResponseGenerator implements RetrieveResponseGenerator
		{
			SetOfDicomFiles m_sdf=null;	
			public void close()
			{
				// TODO Auto-generated method stub				
			}

			public SetOfDicomFiles getDicomFiles()
			{
				return m_sdf;
			}

			public String getErrorComment()
			{
				// TODO Auto-generated method stub
				return null;
			}

			public AttributeTagAttribute getOffendingElement()
			{
				// TODO Auto-generated method stub
				return null;
			}

			public int getStatus()
			{
				// TODO Auto-generated method stub
				return 0;
			}
			private String PseudoUID()
			{
				return 
				new Long(Calendar.getInstance().getTimeInMillis()).toString()+
				new Integer(Calendar.getInstance().get(Calendar.DATE)).toString()+"."+
				new Integer(Calendar.getInstance().get(Calendar.YEAR)).toString()+"."+
				new Long(m_RandSeed++).toString();				
			}
			private void WriteFromStream(String fname, InputStream is)
			throws FileNotFoundException, IOException
			{
				File f= new File(fname);
				FileOutputStream fos=new FileOutputStream(fname);
				final byte[] buf=new byte[16384];
				int bytesRead=0;
				while((bytesRead=is.read(buf))>0)
				{
					fos.write(buf, 0, bytesRead);
				}
				fos.flush();
				fos.close();
			}
			
			private String SaveDicomFile(InputStream is, AttributeList request)
			throws IOException, DicomException
			{
				AttributeList fal=new AttributeList();
				String fname=(m_StoreFolder+"/"+PseudoUID()).replace('\\', '/').replace("//", "/");
				fal.read(new DicomInputStream(is));
				fal.putAll(request);
				fal.removeGroupLengthAttributes();
				fal.removeMetaInformationHeaderAttributes();
				fal.remove(TagFromName.DataSetTrailingPadding);
				fal.correctDecompressedImagePixelModule();
				FileMetaInformation.addFileMetaInformation(fal, 
						TransferSyntax.ExplicitVRLittleEndian,"SERVERTEST");
				fal.write(fname,TransferSyntax.ExplicitVRLittleEndian,true,true);
				new File(fname).deleteOnExit();
				return fname;				
			}
			
			private String DownloadAndSaveDicomFile(XNATRestAdapter xre,String URI, AttributeList request)
			{
				HttpMethodBase get;
				String path=(m_XNATServer+"/"+URI).replace("//", "/");
				Tools.LogMessage(Priority.INFO, "Retrieving file "+path);
				path=URI.substring(5);
				get=xre.PerformConnection(XNATRestAdapter.GET, path, "");
				if(get==null) return null;
				try
				{
					return SaveDicomFile(get.getResponseBodyAsStream(),request);
				}
				catch(Exception e)
				{
					return null;
				}
				finally
				{
					get.releaseConnection();
				}
			}
			byte[] expandBuf(byte[] buf)
			throws OutOfMemoryError
			{
				if(buf.length<1) return buf;
				byte[] newBuf=new byte[buf.length*2];
				for(int i=0; i<buf.length; i++)
					newBuf[i]=buf[i];
				return newBuf;
			}
			
			byte[] readZipBuf(ZipInputStream zin)
			{
				ByteArrayOutputStream baos=new ByteArrayOutputStream();
				try
				{
					int len=0;
					byte[] buf=new byte[2048];
					while((len=zin.read(buf))>0)
					{
						baos.write(buf,0,len);
					}
					return baos.toByteArray();
				}
				catch(Exception e)
				{
					return null;
				}
				finally
				{
					if(baos!=null) try{ baos.close();}catch(Exception e){}
				}				
			}
					
			private void retrieveSeries(AttributeList al, boolean bAppend, boolean bUseCompression)
			{
//				bUseCompression=false;//??
				System.gc();
				XNATRestAdapter xre = new XNATRestAdapter(m_XNATServer,m_XNATUser,m_XNATPass);			
				HttpMethodBase get;
								
//				AttributeList mod_request=new AttributeList();
//				mod_request.putAll(request);
				XNATQueryGenerator.GetVocabulary().modifySOPInstUID(al, false);
				
				String path=XNATQueryGenerator.getRetrieveRESTQuery(InformationEntity.SERIES, al);
				Tools.LogMessage(Priority.INFO, "Rest query path: "+path);
				if(path==null)
				{
					Tools.LogMessage(Priority.ERROR, "Error generating REST retrieve query");
					return;
				}
				if(bUseCompression)
					path+=(path.contains("?")?"&":"?")+"format=zip";				
				if(null==(get=xre.PerformConnection(XNATRestAdapter.GET, path, ""))) return;				
/*				
				try
				{
//					String resp=get.getResponseBodyAsString();
//					System.err.println(resp);
				}
				catch(Exception e){return;}
*/				
				XNATQueryGenerator.GetVocabulary().modifySOPInstUID(al, true);

				try
				{
					//parse xml table and download each individual files
					if(!bUseCompression)
					{
						//2. parse the response
						LinkedList<TreeMap<String,String>> row_map=
							XNATTableParser.GetRows(new SAXReader().read(get.getResponseBodyAsStream()),true,"header");//bGenSearch?"header":null);
						String URI;
						//get the actual DICOM files!
						if(!bAppend || m_sdf==null) m_sdf=new SetOfDicomFiles();
						for(TreeMap<String,String> row:row_map)
						{
							URI=row.get("URI");
							if(URI==null) continue;
							String fname=DownloadAndSaveDicomFile(xre,URI,al);
							if(fname!=null)
								m_sdf.add(fname);
						}
	//					if(m_sdf.size()<1) m_sdf=null;
					}
					//download entire file archive.
					else
					{
						if(!bAppend || m_sdf==null) m_sdf=new SetOfDicomFiles();						
						ZipInputStream zin=new ZipInputStream(get.getResponseBodyAsStream());
						String fname=null;
						ZipEntry entry;
						while((entry=zin.getNextEntry())!=null)
						{
							try
							{
								byte[] buf=readZipBuf(zin);
								if(buf==null || buf.length<1) 
								{
									continue;
								}
								fname=SaveDicomFile(new ByteArrayInputStream(buf),al);
							}
							catch(IOException ioex){continue;}
							finally{zin.closeEntry();}
							
							if(fname!=null) m_sdf.add(fname);
						}
						zin.close();
					}
				}
				catch(Exception e)
				{
					Tools.LogException(Priority.ERROR, "Error parsing the response to REST query", e);
					m_sdf=null;
				}
				finally
				{
					get.releaseConnection();
				}
			}
			private boolean HasUniqueID(AttributeList al, InformationEntity level)
			{
				if(level.compareTo(InformationEntity.SERIES)==0)
				{
					Attribute at=al.get(new AttributeTag(0x0020,0x000E));
					if(at==null || at.getSingleStringValueOrNull()==null) return false;
					return true;
				}
				else if(level.compareTo(InformationEntity.STUDY)==0)
				{
					Attribute at=al.get(new AttributeTag(0x0020,0x000D));
					if(at==null || at.getSingleStringValueOrNull()==null) return false;
					return true;
				}
				return false;
			}
			
			public void performRetrieve(String retrieveSOPClassUID,
					AttributeList request, boolean relational)
			{
				String qLevel=Attribute.getSingleStringValueOrNull(request,TagFromName.QueryRetrieveLevel);
				InformationEntity ieWanted  = getInformationEntityForQueryRetieveLevel(qLevel);
				Tools.LogMessage(Priority.INFO, ieWanted.toString()+" retrieve request received");
				boolean bStudyDefined=HasUniqueID(request,InformationEntity.STUDY),
						bSeriesDefined=HasUniqueID(request,InformationEntity.SERIES);

				//first, update the attribute list with subject info.
				XNATRestAdapter xra = new XNATRestAdapter(m_XNATServer,m_XNATUser,m_XNATPass);			
				HttpMethodBase get;
				String p="experiments?ID="+XNATQueryGenerator.GetValueFromAttributeList("stinstuid", request)+
					"&columns=ID,project,label,subject_ID,subject_label&format=xml";
				get=xra.PerformConnection(XNATRestAdapter.GET, p, "");
				try
				{
					String rsp=get.getResponseBodyAsString();
					LinkedList<TreeMap<String,String>> rm=
						XNATTableParser.GetRows(new SAXReader().read(get.getResponseBodyAsStream()),true,"header");//bGenSearch?"header":null);
					int i=0;
					for(TreeMap<String,String> row:rm)
					{
						if(i>0) break;
						AttributeList received=XNATQueryGenerator.GetVocabulary().GetDicomEntry(row,ieWanted);
						//patient name, id and staccessionnum
						request.put(received.get(new AttributeTag(0x0010,0x0010)));
						request.put(received.get(new AttributeTag(0x0010,0x0020)));
						request.put(received.get(new AttributeTag(0x0008,0x0050)));
						i++;
					}
					get.releaseConnection();
				}
				catch(Exception e){}																
				
				//then, proceed with series retrieve.
				if(ieWanted.compareTo(InformationEntity.SERIES)==0 && bSeriesDefined)
				{
					retrieveSeries(request,false,true); //??
				}
				else if((ieWanted.compareTo(InformationEntity.STUDY)==0 && bStudyDefined)
						|| (ieWanted.compareTo(InformationEntity.SERIES)==0 && !bSeriesDefined))
				{
					m_sdf=null;
					String path=XNATQueryGenerator.getRESTQuery(InformationEntity.SERIES, request);
					Tools.LogMessage(Priority.INFO, "REST query: "+path);
					
					if(path==null)
					{
						Tools.LogMessage(Priority.ERROR, "Error generating REST query");
						return;
					}
					//query for all series.
					XNATRestAdapter xre=new XNATRestAdapter(m_XNATServer,m_XNATUser,m_XNATPass);
					HttpMethodBase method=xre.PerformConnection(XNATRestAdapter.GET, path, "");
					
					if(method==null) return;
					try
					{
						String resp=method.getResponseBodyAsString();
						Tools.LogMessage(Priority.INFO, "REST query response: "+ resp);
					}
					catch(Exception e){return;}
					method.releaseConnection();
					try
					{
						//2. parse the response				    			- using XND's classes
						LinkedList<TreeMap<String,String>> row_map=
							XNATTableParser.GetRows(new SAXReader().read(method.getResponseBodyAsStream()),true,"header");//bGenSearch?"header":null);
						for(TreeMap<String,String> row:row_map)
						{
							//3. translate the response to DICOM's AttributeList 
							AttributeList al=new AttributeList();
							al.putAll(request);
							AttributeList received=XNATQueryGenerator.GetVocabulary().GetDicomEntry(row,InformationEntity.SERIES);
							al.putAll(received);
							XNATQueryGenerator.GetVocabulary().modifySOPInstUID(al, true);
							if(al.size()>0)
							{
								retrieveSeries(al,true,true);
							}
						}
					}
					catch(Exception e)
					{
						Tools.LogException(Priority.ERROR, "Error parsing the response to REST query", e);
					}					
				}
				else m_sdf=null;
				if(m_sdf!=null) //update cache 
				{
					UpdateCache(m_sdf);
				}
			}
		}
		
		public RetrieveResponseGenerator newInstance()
		{
			return new TestRetrieveResponseGenerator();
		}
	};
	
	private class XNATQueryResponseGeneratorFactory implements QueryResponseGeneratorFactory
	{
//		private DicomDictionary m_dictionary = new DicomDictionary();
		
		class TestResponseGenerator implements QueryResponseGenerator
		{
			LinkedList<AttributeList> m_curLst=new LinkedList<AttributeList>();
			private int m_status = ResponseStatus.Success;
			private String m_qLevel=null;
			private String m_ErrComment=null;
			
			public TestResponseGenerator ()
			{
				
			}
			public boolean allOptionalKeysSuppliedWereSupported()
			{
				// TODO Auto-generated method stub
				return false;
			}

			public void close()
			{
				// TODO Auto-generated method stub
				
			}

			public String getErrorComment()
			{
				// TODO Auto-generated method stub
				return m_ErrComment;
			}

			public AttributeTagAttribute getOffendingElement()
			{
				// TODO Auto-generated method stub
				return null;
			}

			public int getStatus()
			{
				return m_status;
			}

			public AttributeList next()
			{				
				if(m_curLst.size()>0)
				{
					AttributeList lst=m_curLst.getFirst();
					m_curLst.removeFirst();
					return lst;
				}
				else return null;
			}
			private void setErrorStatus(int status, String comment)
			{
				m_status=status;
				m_ErrComment = comment;
			}
			public void performQuery(String querySOPClassUID,
					AttributeList queryIdentifier, boolean relational)
			{
				m_qLevel=Attribute.getSingleStringValueOrNull(queryIdentifier,TagFromName.QueryRetrieveLevel);
				InformationEntity ieWanted  = getInformationEntityForQueryRetieveLevel(m_qLevel);

				//1. perform POST request via REST API - using XND's classes
				//XNATRestAdapter xre = new XNATRestAdapter("http://nilpc190:7080/cnda_xnat","mmilch","abc123");				
				//XNATRestAdapter xre = new XNATRestAdapter("https://cndabeta.wustl.edu","mmilch","abc123");				
				//XNATRestAdapter xre = new XNATRestAdapter("http://central.xnat.org","mmilch","abc123");				
				XNATRestAdapter xre = new XNATRestAdapter(m_XNATServer,m_XNATUser,m_XNATPass);
				
				boolean bGenSearch=false;
				HttpMethodBase method;
				if(ieWanted.compareTo(InformationEntity.PATIENT)==0)
//				if(ieWanted.compareTo(InformationEntity.STUDY)<=0)
				{
					//now, for Patient/Study level, use generic xml search.
					String xml=XNATQueryGenerator.getQueryXML(ieWanted, querySOPClassUID, queryIdentifier);
					//verify the query xml
					Tools.LogMessage(Priority.INFO, "XNAT Search engine query xml body:\n"+xml);
					if(!xre.VerifyConnection()) return;
					bGenSearch=true;
					method=xre.PerformConnection(XNATRestAdapter.POST, "/search", xml);
				}
				else if(ieWanted.compareTo(InformationEntity.SERIES)<=0)
//				else if(ieWanted.compareTo(InformationEntity.SERIES)==0)
				{
					String path=XNATQueryGenerator.getRESTQuery(ieWanted, queryIdentifier);
					Tools.LogMessage(Priority.INFO, "REST query string:\n"+path);
					if(path==null)
					{
						Tools.LogMessage(Priority.ERROR, "Error generating REST query");
						return;
					}
					method=xre.PerformConnection(XNATRestAdapter.GET, path, "");					
				}
				else return;	
				if(method==null) return;
/*
				try
				{
//					String resp=method.getResponseBodyAsString();
//					System.err.println(resp);
				}				
				catch(Exception e){return;}
				method.releaseConnection();
*/				
				try
				{
					//2. parse the response				    			- using XND's classes
					LinkedList<TreeMap<String,String>> row_map=
						XNATTableParser.GetRows(new SAXReader().read(method.getResponseBodyAsStream()),!bGenSearch,"header");//bGenSearch?"header":null);
					for(TreeMap<String,String> row:row_map)
					{
						//3. translate the response to DICOM's AttributeList
						AttributeList al=new AttributeList();
						al.putAll(queryIdentifier);
						AttributeList received=XNATQueryGenerator.GetVocabulary().GetDicomEntry(row,ieWanted);
						al.putAll(received);
						XNATQueryGenerator.GetVocabulary().modifySOPInstUID(al, true);
						if(al.size()>0)
							m_curLst.add(al);
					}
				}
				catch(Exception e)
				{
					Tools.LogException(Priority.ERROR, "Error parsing the response to REST query", e);
				}
				finally
				{
					method.releaseConnection();
				}
			}
		}
		
		public QueryResponseGenerator newInstance()
		{
			return new TestResponseGenerator();
		}
	}
}