package org.nrg.xnat.gateway;

import java.io.File;
import java.util.Iterator;

import org.nrg.xdat.bean.XdatCriteriaSetBean;
import org.nrg.xdat.bean.XdatSearchFieldBean;
import org.nrg.xdat.bean.XdatStoredSearchBean;

import com.pixelmed.dicom.Attribute;
import com.pixelmed.dicom.AttributeList;
import com.pixelmed.dicom.AttributeTag;
import com.pixelmed.dicom.DicomDictionary;
import com.pixelmed.dicom.InformationEntity;

public class XNATQueryGenerator
{
	private static XNATVocabulary m_vcbl=null;
	private static DicomDictionary m_dd=new DicomDictionary();
	
	public static void LoadVocabulary(String fname)
	{
		m_vcbl=new XNATVocabulary(new File(fname));
	}
	public static XNATVocabulary GetVocabulary(){return m_vcbl;}
	private static String GetRootStoredSearchElementName(InformationEntity ie)
	{
		String res="";
		if (ie == InformationEntity.PATIENT)		res="xnat:subjectData";
		else if (ie == InformationEntity.STUDY)		res="xnat:mrsessionData";		
		else if (ie == InformationEntity.SERIES)	res="xnat:scanData";
		else return null;
		return res;
	}
	public static String getRetrieveRESTQuery(InformationEntity ie, AttributeList al)
	{
		if(ie.compareTo(InformationEntity.SERIES)==0)
		{
			//get series path
			String path=getRESTQuery(ie,al);
			String scanID=GetValueFromAttributeList("id",al);
			if(scanID==null) return null;
			path+="/"+scanID+"/files";			
			return path;
		}
		return null;
	}
	
	public static String getRESTQuery(InformationEntity ie, AttributeList query)
	{
		if(ie.compareTo(InformationEntity.SERIES)==0)
		{
			String path="/experiments/";
			//get experiment ID
			String expID=GetValueFromAttributeList("stinstuid",query);
			if(expID==null) return null;
			path+=expID+"/scans";
			return path;
		}
		else if (ie.compareTo(InformationEntity.STUDY)==0)
		{
			String path="/experiments";
			String expID=GetValueFromAttributeList("stinstuid",query);
			String label=GetValueFromAttributeList("staccessionnum",query);
			String modality=GetValueFromAttributeList("stmodality",query);
			String date=GetValueFromAttributeList("stdate",query);
			String patname=GetValueFromAttributeList("patname",query);
			String patid=GetValueFromAttributeList("patid",query);
			boolean bFirst=true;
			
			if(expID!=null)		{path+=(bFirst?"?":"&")+"ID="+expID;bFirst=false;}
			if(label!=null)		{path+=(bFirst?"?":"&")+"label="+label;bFirst=false;}
			if(modality!=null)	{path+=(bFirst?"?":"&")+"xsiType="+modality;bFirst=false;}
			if(date!=null)		{path+=(bFirst?"?":"&")+"date="+date;bFirst=false;}
			if(patname!=null)	{path+=(bFirst?"?":"&")+"subject_label="+patname;bFirst=false;}
			if(patid!=null)		{path+=(bFirst?"?":"&")+"subject_ID="+patid;bFirst=false;}
			return path;
		}
		return null;
	}
	
	public static String GetValueFromAttributeList(String xnat_field_id, AttributeList al)
	{
		XNATVocabularyEntry xve;
		//get experiment ID
		xve=m_vcbl.GetXNATEntry(xnat_field_id);
		if(xve==null) return null;
		Attribute attr=al.get(xve.m_DICOMTag);
		if(attr==null) return null;
		return XNATVocabulary.dcmToXNATField(attr.getSingleStringValueOrNull(),xnat_field_id);
	}
	
	public static String getQueryXML(InformationEntity ie, String querySOPClassUID, AttributeList query)
	{
		XdatStoredSearchBean search = new XdatStoredSearchBean();
		String name=GetRootStoredSearchElementName(ie);
		if(name==null) return null;
		search.setRootElementName(name);
		XdatSearchFieldBean sf;
		for (Iterator i=query.values().iterator(); i.hasNext();)
		{
			Attribute a = (Attribute)i.next();
			AttributeTag at = a.getTag();
			InformationEntity ieAttribute = m_dd.getInformationEntityFromTag(at);
			if(ieAttribute==null) continue;
			//add fields
			if(ieAttribute.compareTo(ie)<=0)
			{
				XNATVocabularyEntry xve=m_vcbl.GetDcmEntry(at);
				if(xve!=null)
				{
					sf=new XdatSearchFieldBean();
					sf.setElementName(xve.m_elementName);
					sf.setFieldId(xve.m_field_id);
					sf.setHeader(xve.m_search_column_alias);
					sf.setType(xve.m_type);
					sf.setSequence(0);
					search.addSearchField(sf);
					
					//try to add a where clause
					String val=a.getSingleStringValueOrDefault(null);
					if(val!=null)
					{
						XdatCriteriaSetBean xcsb=xve.getCriteriaSet("%"+val+"%");
						if(xcsb!=null) search.addSearchWhere(xcsb);
					}
				}
			}
		}
		return search.toString();
	}	
}
