package org.nrg.xnat.repository;

import java.io.File;
import java.util.Collection;
import java.util.TreeMap;
import java.util.TreeSet;
import java.util.Vector;

import javax.swing.filechooser.FileSystemView;

import org.nrg.xnat.desktop.model.CElement;
import org.nrg.xnat.desktop.utils.FSObject;
import org.nrg.xnat.desktop.utils.Utils;


public class ItemRecord implements Comparable
{	
	protected TreeMap<String,ItemTag> m_Tags=new TreeMap<String,ItemTag>();
	protected String m_AbsolutePath = "";
	protected String m_RelativePath = "";
	
	public boolean IsCollectionDefined()
	{
		return TagByName("Collection_ID")!=null;
	}
	public String GetCollectionName()
	{
		final String[] def_tags={"Project","Subject","Experiment","Scan"};
		String tval;
		String r="";
		for(String s:def_tags)
		{
			tval=this.GetTagValue(s);
			if(tval==null) tval="";
			if(r.length()>0) r=r+"_"+tval;
			else r=tval;
		}
		return r;
	}
	public String GetColID()
	{
		return this.GetTagValue("Collection_ID");
	}
/*	
	public void AddResource(String rel_path)
	{
//		m_Resources.add(rel_path);
		ItemTag tag;
		if(m_Tags.containsKey("__resources"))
		{
			tag=m_Tags.get("__resources");
			tag.AddValue(rel_path);
		}
		else
		{
			tag=new ItemTag("__resources",rel_path,true);
			m_Tags.put(tag.GetName(), tag);
		}
	}

	public String[] GetAllResources()
	{
		try
		{
			return m_Tags.get("__resources").GetAllValues();
		}
		catch(Exception e)
		{
			return new String[0];
		}
	}
	public void RemoveResource(String rel_path)
	{
		try
		{
			m_Tags.get("__resources").RemoveValue(rel_path);
		}
		catch(Exception e){}
	}
	public int CountResources()
	{
		try
		{
			return 
				m_Tags.get("__resources").GetAllValues().length;
		}
		catch(Exception e){return 0;}
	}
*/	
	public int compareTo(Object o)
	{
		if(m_AbsolutePath!=null)
			return m_AbsolutePath.compareTo(((ItemRecord)o).GetAbsolutePath());
		if(m_RelativePath!=null) 
			return m_RelativePath.compareTo(((ItemRecord)o).GetRelativePath());
		else return 0;
	}
	
	public FSObject[] ListFiles(boolean bCheckDirs)
	{
		if(m_AbsolutePath==null || m_AbsolutePath.length()<1) return new FSObject[0];
		File f=new File(m_AbsolutePath);
//		File[] res;
		try
		{
//			res=f.listFiles();
			return Utils.m_nfm.ListFiles(f,bCheckDirs);
		}
		catch(Exception e)
		{
			return new FSObject[0];
		}
	}
	public File GetFile()
	{
		if(m_AbsolutePath!=null && m_AbsolutePath.length()>0) return new File(m_AbsolutePath);
		return null;
	}
	
	public String GetFileParent()
	{
		File f=GetFile();
		if(f==null) return "";
		return f.getParent();
	}
	
	public String GetFileName()
	{
		File f=GetFile();
		if(f==null && m_RelativePath!=null) 
			f=new File(m_RelativePath);
		if(f!=null) return f.getName();
		return null;
	}
	
	public ItemRecord(ItemRecord ir)
	{
		m_Tags.putAll(ir.m_Tags);
		m_AbsolutePath=ir.m_AbsolutePath;
		m_RelativePath=ir.m_RelativePath;
	}
	public String PrintTags()
	{
		ItemTag[] tags=GetAllTags();
		String s="";
		if(tags.length<1) return s;
		for(int i=0; i<tags.length; i++)
		{
			s=s+tags[i].PrintTag()+"\n";
		}
		return s;
	}
	public String GetAbsolutePath()
	{
		return m_AbsolutePath;
	}
	public String GetRelativePath()
	{
		if(m_RelativePath==null) return null;
		while (m_RelativePath.startsWith("/")) m_RelativePath=m_RelativePath.substring(1);
		return m_RelativePath.replace('\\', '/');
	}
	
	public ItemRecord(String abs_path, String rel_path)
	{
		m_AbsolutePath=abs_path;
		m_RelativePath=rel_path;
	}
	/**
	 * @param abs_path <prefix>/<root_fold>/<suffix>/<file>
	 * @param root	   <prefix>/<root_fold>
	 * @param bInclRoot 
	 * @return <root_fold>/<suffix>/<file>, if bInclRoot=true; <suffix>/<file>, otherwise
	 *		
	 */
	public static String RelativeFromAbsolute(String abs_path, String root, boolean bInclRoot)
	{
		String res;
		if(abs_path!=null && root!=null && root.length()>0 && abs_path.startsWith(root))
		{
			res=abs_path.substring(root.length());
			while((res.startsWith("/") || res.startsWith("\\")))
				res=res.substring(1);
		}		
		else res=null;
		if(!bInclRoot) return res;		
		
		//add root
		if(res!=null)
		{
			if(res.length()>0)
				return (new File(root).getName()+"/"+res).replace('\\','/');
			else return new File(root).getName().replace('\\','/');
		}
		return res;
	}
	public ItemTag TagByName(String name)
	{
		if(m_Tags.containsKey(name))
			return m_Tags.get(name);
		else return null;
	}
	public String GetTagValue(String name)
	{
		if(m_Tags.containsKey(name))
			return m_Tags.get(name).GetFirstValue();
		else return null;
	}
	public void TagSet(ItemTag tag)
	{
		if(tag==null) return;
		if(m_Tags.containsKey(tag.GetName())) m_Tags.remove(tag.GetName());
		m_Tags.put(tag.GetName(),tag);
	}
	public void TagsMerge(Collection<ItemTag> tags)
	{
		ItemTag lit;
		for(ItemTag it:tags)
		{
			lit=m_Tags.get(it.GetName());
			if(lit==null)
			{
				m_Tags.put(it.GetName(), it);
			}
			else
			{
				lit.SetAllowMultiple(true);
				lit.AddValues(it.GetValues());
			}
		}
	}
	public void TagsSet(Collection<ItemTag> tags)
	{
		m_Tags.clear();
		for(ItemTag it:tags) m_Tags.put(it.GetName(), it);
	}
	public void SetAbsolutePath(String new_path)
	{
		m_AbsolutePath=new_path;
	}
	public void TagsSet(ItemTag[] tags)
	{
		for(int i=0; i<tags.length; i++) m_Tags.put(tags[i].GetName(), tags[i]);
	}
	public void TagRemove(String name)
	{
		m_Tags.remove(name);
	}
	public void RemoveAllTags(){m_Tags.clear();}
	public ItemTag GetTagByName(String name)
	{
		return m_Tags.get(name);
	}
	public Collection<ItemTag> GetTagCollection()
	{
		return m_Tags.values();
	}
	public ItemTag[] GetAllTags()
	{
		return m_Tags.values().toArray(new ItemTag[0]);
	}
	public ItemTag GetFirstTag()
	{
		if(m_Tags.size()<1) return null;
		return m_Tags.get(m_Tags.firstKey());
	}
	public boolean Update(ItemRecord template)
	{
		TagSet(template.GetFirstTag());
		return true;
	}	
}